import datetime
import json
import logging
import os
import re
from typing import Optional

import pandas as pd
import pkg_resources
import requests
import urllib3
from urllib3.connectionpool import MaxRetryError, SSLError

from seeq import spy
from seeq.base import system, util
from seeq.sdk import *
from seeq.sdk.rest import ApiException
from seeq.spy import _common
from seeq.spy._config import Setting
from seeq.spy._errors import *
from seeq.spy._session import Session
from seeq.spy._status import Status

AUTOMATIC_PROXY_DETECTION = '__auto__'


def login(username=None, password=None, *, access_key=None, url=None, directory='Seeq',
          ignore_ssl_errors=False, proxy=AUTOMATIC_PROXY_DETECTION, credentials_file=None, force=True,
          quiet=False, status=None, session: Session = None, private_url=None, auth_token=None, csrf_token=None):
    """
    Establishes a connection with Seeq Server and logs in with a set of
    credentials. At least one set of credentials must be provided.
    Applicable credential sets are:
        - username + password (where username is in "Seeq" user directory)
        - username + password + directory
        - access_key + password
        - credentials_file (where username is in "Seeq" user directory)
        - credentials_file + directory

    Parameters
    ----------
    username : str, optional
        Username for login purposes. See credentials_file argument for
        alternative.

    password : str, optional
        Password for login purposes. See credentials_file argument for
        alternative.

    access_key: str, optional
        Access Key for login purposes. Access Keys are created by individual
        users via the Seeq user interface in the upper-right user profile
        menu. An Access Key has an associated password that is presented
        to the user (once!) upon creation of the Access Key, and it must be
        supplied via the "password" argument. The "directory" argument must
        NOT be supplied.

    url : str, default 'http://localhost:34216'
        Seeq Server url. You can copy this from your browser and cut off
        everything to the right of the port (if present).
        E.g. https://myseeqserver:34216

    directory : str, default 'Seeq'
        The authentication directory to use. You must be able to supply a
        username/password, so some passwordless Windows Authentication
        (NTLM) scenarios will not work. OpenID Connect is also not
        supported. If you need to use such authentication schemes, set up
        a Seeq Data Lab server.

    ignore_ssl_errors : bool, default False
        If True, SSL certificate validation errors are ignored. Use this
        if you're in a trusted network environment but Seeq Server's SSL
        certificate is not from a common root authority.

    proxy : str, default '__auto__'
        Specifies the proxy server to use for all requests. The default
        value is "__auto__", which examines the standard HTTP_PROXY and
        HTTPS_PROXY environment variables. If you specify None for this
        parameter, no proxy server will be used.

    credentials_file : str, optional
        Reads username and password from the specified file. If specified, the
        file should be plane text and contain two lines, the first line being
        the username, the second being the user's password.

    force : str, default True
        If True, re-logs in even if already already logged in. If False, skips
        login if already logged in. You should include a spy.login(force=False)
        cell if you are creating example notebooks that may be used in Jupyter
        environments like Anaconda, AWS SageMaker or Azure Notebooks.)

    quiet : bool, default False
        If True, suppresses progress output. Note that when status is
        provided, the quiet setting of the Status object that is passed
        in takes precedent.

    status : spy.Status, optional
        If supplied, this Status object will be updated as the command
        progresses.

    session : spy.Session, optional
        If supplied, the Session object (and its Options) will be used to
        store the login session state. This is useful to log in to different
        Seeq servers at the same time or with different credentials.

    private_url : str
        If supplied, this will be the URL used for communication with the Seeq
        Server API over private networks.  Generally for internal use only.

    auth_token : str
        Private argument for Data Lab use only.

    csrf_token : str
        Private argument for Data Lab use only.

    Examples
    --------
    Log in to two different servers at the same time:

    >>> session1 = Session()
    >>> session2 = Session()
    >>> spy.login(url='https://server1.seeq.site', username='mark', password='markpassword', session=session1)
    >>> spy.login(url='https://server2.seeq.site', username='alex', password='alexpassword', session=session2)

    """
    _common.validate_argument_types([
        (username, 'username', str),
        (password, 'password', str),
        (access_key, 'access_key', str),
        (url, 'url', str),
        (private_url, 'private_url', str),
        (directory, 'directory', str),
        (ignore_ssl_errors, 'ignore_ssl_errors', bool),
        (proxy, 'proxy', str),
        (credentials_file, 'credentials_file', str),
        # Note: Although auth_token is no longer a supported authentication method for non-Seeq Data Lab scenarios,
        # it is still used by Seeq Data Lab code to log the user in.
        (auth_token, 'auth_token', str),
        (csrf_token, 'csrf_token', str),
        (force, 'force', bool),
        (quiet, 'quiet', bool),
        (status, 'status', Status),
        (session, 'session', Session)
    ])

    status = Status.validate(status, quiet)
    session = Session.validate(session)

    if private_url is not None and url is None:
        raise SPyValueError('private_url argument cannot be specified without also specifying url argument')

    try:
        _login(username, password, access_key, url, directory, ignore_ssl_errors, proxy,
               credentials_file, auth_token, private_url, force, status, csrf_token, session)
    except SPyException as e:
        status.update(str(e), Status.FAILURE)
        raise

    user_string = session.user.username
    user_profile = ''
    if session.user.first_name:
        user_profile = session.user.first_name
    if session.user.last_name:
        user_profile += ' ' + session.user.last_name
    if session.user.is_admin:
        user_profile += ' [Admin]'
    if len(user_profile) > 0:
        user_string += ' (%s)' % user_profile.strip()

    status.update(f'Logged in to <strong>{session.public_url}</strong> '
                  f'successfully as <strong>{user_string}</strong>.\n'
                  f'Seeq Server Version: <strong>{spy.server_version}</strong>\n'
                  f'Seeq Python Module Version: <strong>{spy.__version__}</strong>',
                  Status.SUCCESS)


def _login(username, password, access_key, url, directory, ignore_ssl_errors, proxy, credentials_file, auth_token,
           private_url, force, status, csrf_token, session: Session):
    if access_key and username:
        raise SPyValueError('"username" argument must be omitted when supplying "access_key" argument')

    if access_key and directory != 'Seeq':
        raise SPyValueError('"directory" argument must be omitted when supplying "access_key" argument')

    try:
        if force or not session.client:
            # Clear out any global state before attempting to log in
            _clear_login_state(quiet=True, status=None, session=session)

            if url:
                session.public_url = url
                session.private_url = url
            else:
                session.public_url = Setting.SEEQ_URL.get() if Setting.SEEQ_URL.get() else 'http://localhost:34216'
                session.private_url = Setting.PRIVATE_URL.get() if Setting.PRIVATE_URL.get() else session.public_url

            if private_url:
                # User is overriding the private_url that was "calculated" in logic above
                session.private_url = private_url

            _client_login(auth_token, credentials_file, directory, ignore_ssl_errors, password, proxy, status,
                          session.get_api_url(), username, access_key, csrf_token, session)

        system_api = SystemApi(session.client)
        server_status = system_api.get_server_status()  # type: ServerStatusOutputV1
        session.server_version = server_status.version
        validate_seeq_server_version(
            session,
            status,
            # We allow a version mismatch here because in the case of Data Lab, this login call happens when the
            # kernel is initialized, and if an exception is thrown, it is not seen by the user. Then subsequent calls to
            # other functions like spy.search() fail with a "not logged in" error instead of failing with a version
            # mismatch error. That's confusing. So instead, we just warn during login (but succeed) and then fail later
            # when other functions are called.
            allow_version_mismatch=True)

        validate_data_lab_license(session)

        users_api = UsersApi(session.client)
        session.user = users_api.get_me()

        folders_api = FoldersApi(session.client)

        # noinspection PyBroadException
        try:
            session.corporate_folder = folders_api.get_folder(folder_id='corporate')
        except BaseException:
            # This can happen in cases where the user does not have access rights to the Corporate folder
            session.corporate_folder = None

    except BaseException:
        logout(quiet=True, session=session)
        raise


def _client_login(auth_token, credentials_file, directory, ignore_ssl_errors, password, proxy, status, api_client_url,
                  username, access_key, csrf_token, session: Session):
    # Annoying warnings are printed to stderr if connections fail
    logging.getLogger("requests").setLevel(logging.FATAL)
    logging.getLogger("urllib3").setLevel(logging.FATAL)
    urllib3.disable_warnings()

    cert_file = Setting.get_seeq_cert_path()
    if cert_file and os.path.exists(cert_file):
        session.client_configuration.set_certificate_path(cert_file)

    key_file = Setting.get_seeq_key_path()
    if key_file and os.path.exists(key_file):
        session.client_configuration.key_file = key_file

    session.client_configuration.verify_ssl = not ignore_ssl_errors

    if proxy == AUTOMATIC_PROXY_DETECTION:
        if api_client_url.startswith('https') and 'HTTPS_PROXY' in os.environ:
            session.client_configuration.proxy = os.environ['HTTPS_PROXY']
        elif 'HTTP_PROXY' in os.environ:
            session.client_configuration.proxy = os.environ['HTTP_PROXY']
    elif proxy is not None:
        session.client_configuration.proxy = proxy

    _client = ApiClient(api_client_url, configuration=session.client_configuration)
    _client.set_default_header('x-sq-origin', 'Data Lab')

    auth_api = AuthApi(_client)
    directories = dict()
    try:
        auth_providers_output = auth_api.get_auth_providers()  # type: AuthProvidersOutputV1
    except MaxRetryError as e:
        if isinstance(e.reason, SSLError):
            raise SPyRuntimeError(f'SSL certificate error. If you trust your network, you can add '
                                  f'the spy.login(ignore_ssl_errors=True) argument.\n\nMore info:\n{e.reason}')

        raise SPyRuntimeError(
            '"%s" could not be reached. Is the server or network down?\n%s' % (api_client_url, e))

    session.auth_providers = auth_providers_output.auth_providers
    for datasource_output in session.auth_providers:  # type: DatasourceOutputV1
        directories[datasource_output.name] = datasource_output

    if auth_token:
        if username or password or credentials_file:
            raise SPyValueError('username, password and/or credentials_file cannot be provided along with auth_token')

        _client.auth_token = auth_token
        _client.csrf_token = csrf_token
    else:
        auth_input = AuthInputV1()

        if access_key:
            username = access_key

        if credentials_file:
            if username is not None or password is not None or access_key is not None:
                raise SPyValueError('If credentials_file is specified, username, '
                                    'access_key and password must be omitted')

            if not os.path.exists(credentials_file):
                repo_root_dir = system.get_repo_root_dir(windows_long_path=False)
                if repo_root_dir:
                    agent_key_file = os.path.join(system.get_repo_root_dir(windows_long_path=False), 'sq-run-data-dir',
                                                  'keys', 'agent.key')
                    if os.path.exists(agent_key_file):
                        # This is a hack to make it easier to run the Documentation notebooks without having to create a
                        # credentials file. It was too easy to forget to delete the credentials file when you created
                        # it, and it would wind up in the distribution.
                        credentials_file = agent_key_file

            if not os.path.exists(credentials_file):
                raise SPyValueError(f'credentials_file "{credentials_file}" not found')

            try:
                with open(credentials_file) as f:
                    lines = f.readlines()
            except Exception as e:
                raise SPyRuntimeError('Could not read credentials_file "%s": %s' % (credentials_file, e))

            if len(lines) < 2:
                raise SPyRuntimeError('credentials_file "%s" must have two lines: username then password')

            username = lines[0].strip()
            password = lines[1].strip()

        if not username or not password:
            if access_key:
                raise SPyValueError('Both access_key and password must be supplied')
            else:
                raise SPyValueError('Both username and password must be supplied')

        auth_input.username = username
        auth_input.password = password

        status.update('Logging in to <strong>%s</strong> as <strong>%s</strong>' % (
            api_client_url, username), Status.RUNNING)

        if not access_key:
            if directory not in directories:
                raise SPyRuntimeError('directory "%s" not recognized. Possible directory(s) for this server: %s' %
                                      (directory, ', '.join(directories.keys())))

            datasource_output = directories[directory]
            auth_input.auth_provider_class = datasource_output.datasource_class
            auth_input.auth_provider_id = datasource_output.datasource_id

        try:
            auth_api.login(body=auth_input)
        except ApiException as e:
            if e.status == 401:
                if access_key:
                    seeq_module_major, _, _, _ = get_module_version_tuple()
                    if seeq_module_major < 50:
                        raise SPyRuntimeError(
                            f'Access Key "{access_key}" is not valid. Log out of the Seeq user interface  (if logged '
                            'in) and then log back in again to reset Access Key validity. If you are still seeing '
                            'this error after doing so, make sure the "access_key" and "password" arguments are '
                            'correct and match an Access Key that you have created in the Seeq user interface.'
                        )
                    else:
                        raise SPyRuntimeError(
                            f'Access Key "{access_key}" is not valid. Log in to the Seeq user interface to reset its '
                            'validity. If you are still seeing this error after doing so, make sure the "access_key" '
                            'and "password" arguments are correct and match an Access Key that you have created in '
                            'the Seeq user interface.'
                        )
                else:
                    raise SPyRuntimeError(
                        f'"{auth_input.username}" could not be logged in with supplied credentials, check username and '
                        'password.')
            else:
                raise
        except MaxRetryError as e:
            raise SPyRuntimeError(
                f'"{api_client_url}" could not be reached. Is the server or network down?\n{e}')
        except Exception as e:
            raise SPyRuntimeError(
                f'Could not connect to Seeq\'s API at {api_client_url} with login "{auth_input.username}".\n{e}')

    # Now that we have succeeded, set all the session variables
    session.client = _client
    session.https_verify_ssl = not ignore_ssl_errors
    session.https_key_file = key_file
    session.https_cert_file = cert_file


def logout(quiet=False, status=None, session: Session = None):
    """
    Logs you out of your current session.

    Parameters
    ----------

    quiet : bool, default False
        If True, suppresses progress output. Note that when status is
        provided, the quiet setting of the Status object that is passed
        in takes precedent.

    status : spy.Status, optional
        If supplied, this Status object will be updated as the command
        progresses.

    session : spy.Session, optional
        The login session to use for this call. See spy.login() documentation
        for info on how to use a Session object.
    """
    _common.validate_argument_types([
        (quiet, 'quiet', bool),
        (status, 'status', Status),
        (session, 'session', Session)
    ])

    session = Session.validate(session)
    _clear_login_state(quiet, status, session, call_logout=True)


def _clear_login_state(quiet, status, session: Session, call_logout=False):
    """
    Clear out any global state and optionally logs you out of your current session.

    Parameters
    ----------

    quiet : bool
        If True, suppresses progress output. Note that when status is
        provided, the quiet setting of the Status object that is passed
        in takes precedent.

    status : spy.Status, optional
        If supplied, this Status object will be updated as the command
        progresses.

    session : spy.Session, optional
        The login session to use for this call. See spy.login() documentation
        for info on how to use a Session object.

    call_logout : bool
        If True, auth_api.logout is called -- otherwise just the session state
        is cleared
    """

    status = Status.validate(status, quiet)

    if session.client is None:
        status.update('No action taken because you are not currently logged in.', Status.FAILURE)
    else:
        if call_logout:
            auth_api = AuthApi(session.client)
            auth_api.logout()

        session.client.logout()

    session.clear()

    status.update('Logged out.', Status.SUCCESS)


def find_user(session: Session, query, exact_match=False):
    """
    Finds a user by using Seeq's user/group autocomplete functionality. Must result in exactly one match or a
    RuntimeError is raised.
    :param session: The login session (necessary to execute this call)
    :param query: A user/group fragment to use for the search
    :param exact_match: If True, it will look for the exact match of the user
    :return: The identity of the matching user.
    :rtype: UserOutputV1
    """
    users_api = UsersApi(session.client)

    if _common.is_guid(query):
        user_id = query
    else:
        identity_preview_list = users_api.autocomplete_users_and_groups(query=query).items
        if len(identity_preview_list) == 0:
            raise SPyRuntimeError('User "%s" not found' % query)
        if len(identity_preview_list) > 1:
            if exact_match:
                identity_preview_list = [x for x in identity_preview_list if x.username == query]
                if len(identity_preview_list) == 0:
                    raise SPyRuntimeError('Exact match for user "%s" not found' % query)
            else:
                raise SPyRuntimeError('Multiple users found that match "%s":\n%s' % (
                    query, '\n'.join([('%s (%s)' % (u.username, u.id)) for u in identity_preview_list])))

        if exact_match and identity_preview_list[0].username != query:
            raise SPyRuntimeError('Exact match for user "%s" not found' % query)

        user_id = identity_preview_list[0].id

    return users_api.get_user(id=user_id)


def find_group(session: Session, query, exact_match=False):
    """
    Finds a group by using Seeq's user/group autocomplete functionality. Must result in exactly one match or a
    RuntimeError is raised.
    :param session: The login session (necessary to execute this call)
    :param query: A user/group fragment to use for the search
    :param exact_match: If True, it will look for the exact match of the user
    :return: The identity of the matching group.
    :rtype: UserGroupOutputV1
    """
    users_api = UsersApi(session.client)
    user_groups_api = UserGroupsApi(session.client)

    if _common.is_guid(query):
        group_id = query
    else:
        identity_preview_list = users_api.autocomplete_users_and_groups(query=query).items
        if len(identity_preview_list) == 0:
            raise SPyRuntimeError('Group "%s" not found' % query)
        if len(identity_preview_list) > 1:
            if exact_match:
                identity_preview_list = [x for x in identity_preview_list if x.name == query]
                if len(identity_preview_list) == 0:
                    raise SPyRuntimeError('Exact match for group "%s" not found' % query)
            else:
                raise SPyRuntimeError('Multiple groups found that match "%s":\n%s' % (
                    query, '\n'.join([('%s (%s)' % (g.name, g.id)) for g in identity_preview_list])))

        if exact_match and identity_preview_list[0].name != query:
            raise SPyRuntimeError('Exact match for user "%s" not found' % query)
        group_id = identity_preview_list[0].id

    return user_groups_api.get_user_group(user_group_id=group_id)


def get_user_timezone(session: Session, default_tz='UTC'):
    """
    Returns the preferred timezone of the user currently logged in, or default_tz if there is no user currently
    logged in.
    :param session: The login session (necessary to fulfill this call).
    :param default_tz: The default timezone to return if no user is logged in.
    :return: The user's preferred timezone, in IANA Time Zone Database format (e.g., 'America/New York')
    :rtype: str
    """
    _common.validate_timezone_arg(default_tz)
    try:
        workbench_dict = json.loads(session.user.workbench)
        return workbench_dict['state']['stores']['sqWorkbenchStore']['userTimeZone']
    except (AttributeError, KeyError, TypeError):
        # This can happen if the user has never logged in interactively (e.g., agent_api_key)
        return default_tz


def get_module_version_tuple():
    match = re.match(r'^(\d+)\.(\d+)\.(\d+)\.(\d+)', spy.__version__)
    seeq_module_major = int(match.group(1))
    seeq_module_minor = int(match.group(2))
    seeq_module_patch = int(match.group(3))
    seeq_module_functional = int(match.group(4))
    return seeq_module_major, seeq_module_minor, seeq_module_patch, seeq_module_functional


def get_server_version_tuple(session: Session):
    match = re.match(r'^R?(?:\d+\.)?(\d+)\.(\d+)\.(\d+)(-v\w+)?(-[-\w]+)?', session.server_version)
    seeq_server_major = int(match.group(1))
    seeq_server_minor = int(match.group(2))
    seeq_server_patch = int(match.group(3))
    return seeq_server_major, seeq_server_minor, seeq_server_patch


def validate_seeq_server_version(session: Session, status: Status, allow_version_mismatch=False):
    seeq_module_major, seeq_module_minor, _, _ = get_module_version_tuple()
    seeq_server_major, seeq_server_minor, seeq_server_patch = get_server_version_tuple(session)

    # The old versioning scheme is like 0.49.3 whereas the new scheme is like 50.1.8
    # See https://seeq.atlassian.net/wiki/spaces/SQ/pages/947225963/Seeq+Versioning+Simplification
    using_old_versioning_scheme = (seeq_server_major == 0)

    message = None

    if using_old_versioning_scheme:
        if seeq_module_major != seeq_server_major or \
                seeq_module_minor != seeq_server_minor:
            message = (f'The major/minor version of the seeq module ({seeq_module_major}.{seeq_module_minor}) '
                       f'does not match the major/minor version of the Seeq Server you are connected to '
                       f'({seeq_server_major}.{seeq_server_minor}) and is incompatible.')
    else:
        if seeq_module_major != seeq_server_major:
            message = (f'The major version of the seeq module ({seeq_module_major}) '
                       f'does not match the major version of the Seeq Server you are connected to '
                       f'({seeq_server_major}) and is incompatible.')

    if message:
        compatible_module_folder = find_compatible_module(session)
        if not compatible_module_folder:
            message += (f'\n\nIt is recommended that you issue the following PIP command to '
                        f'install a compatible version of the seeq module:\n')
            if using_old_versioning_scheme:
                message += f'pip install -U seeq~={seeq_server_major}.{seeq_server_minor}.{seeq_server_patch}'
            else:
                message += f'pip install -U seeq~={seeq_server_major}.{seeq_server_minor}'
        else:
            message += (f'\n\nA compatible module was found here: {compatible_module_folder}\n'
                        f'Consider uninstalling your project-specific seeq module by issuing the following PIP '
                        f'command:\n'
                        f'pip uninstall -y seeq')

        if allow_version_mismatch or session.options.allow_version_mismatch:
            status.warn(message)
        else:
            raise SPyRuntimeError(message)

    return seeq_server_major, seeq_server_minor, seeq_server_patch


def find_compatible_module(session: Session):
    """
    Uses pkg_resources tools to look for a seeq module that is compatible with the version of Seeq Server we're
    connected to. This function is useful in Seeq Data Lab scenarios where the user has installed a "private" version of
    the seeq module (presumably to get a bugfix for SPy) but Seeq Server and Seeq Data Lab have been upgraded in the
    meantime and the user's best course of action is to remove the private version and resume using the "built-in"
    version, which will update with Seeq Data Lab as it is upgraded.
    :return: Path to a compatible module, None if a compatible module is not found.
    """
    seeq_module_major, seeq_module_minor, _, _ = get_module_version_tuple()
    seeq_server_major, seeq_server_minor, seeq_server_patch = get_server_version_tuple(session)

    pkg_env = pkg_resources.Environment()
    seeq_modules = pkg_env['seeq']
    for seeq_module in seeq_modules:  # type: pkg_resources.Distribution
        if seeq_server_major >= 50 and seeq_module.version.startswith(f'{seeq_server_major}.'):
            return seeq_module.location
        elif seeq_server_major < 50 and seeq_module.version.startswith(f'{seeq_server_major}.{seeq_server_minor}.'):
            return seeq_module.location

    return None


def validate_data_lab_license(session: Session):
    system_api = SystemApi(session.client)
    license_status_output = system_api.get_license()  # type: LicenseStatusOutputV1
    for additional_feature in license_status_output.additional_features:  # type: LicensedFeatureStatusOutputV1
        if additional_feature.name == 'Data_Lab':
            if additional_feature.validity == 'Valid':
                return

            raise SPyRuntimeError(f'Seeq Data Lab license is "{additional_feature.validity}", could not log in. '
                                  f'Contact your administrator or speak with Seeq Support at support@seeq.com.')

    raise SPyRuntimeError('Seeq Data Lab is not licensed for this server, could not log in. Contact your administrator '
                          'or speak with Seeq Support at support@seeq.com.')


def validate_login(session: Session, status: Status):
    if session.client is None:
        raise SPyRuntimeError('Not logged in. Execute spy.login() before calling this function.')

    validate_seeq_server_version(session, status)


def localize_datetime(session: Session, dt, default_timezone=None):
    if not pd.isna(dt) and dt.tz is None:
        dt = dt.tz_localize(default_timezone if default_timezone else get_user_timezone(session))
    return dt


def parse_datetime_with_timezone(session: Session, dt: object, nanoseconds_multiplier: int = 1000000,
                                 default_timezone: Optional[object] = None) -> pd.Timestamp:
    if isinstance(dt, pd.Timestamp):
        return localize_datetime(session, dt, default_timezone=default_timezone)

    if isinstance(dt, (float, int)):
        # Assume it's in milliseconds because it came from the front-end
        return pd.Timestamp(int(float(dt) * nanoseconds_multiplier)).tz_localize('UTC')

    if not isinstance(dt, str):
        raise SPyValueError(f'Date/time object {dt} of type "{type(dt)}" not recognized')

    match = re.fullmatch(r'^([^\[\]]+)(\[(.+)])?$', dt)
    if not match or not match.group(3):
        # noinspection PyTypeChecker
        return localize_datetime(session, pd.to_datetime(dt))

    datetime_part = match.group(1)
    tz_part = match.group(3)
    return pd.Timestamp(datetime_part).tz_localize(tz_part)


def validate_start_and_end(session: Session, start, end):
    """
    Takes a pair of start and end times, either of which could be None, and uses some default logic
    to produce a "cleaned up" start and end. For example, if start is after end, then end is reset to
    one hour after start.
    :param session: The login session (necessary to fulfill this call).
    :param start: The start value, or None if you want a default to be calculated.
    :param end: The start value, or None if you want a default to be calculated.
    :return: The calculated (start, end) tuple after having run through the default logic.
    :rtype: tuple
    """
    pd_start = pd.to_datetime(start)  # type: pd.Timestamp
    pd_end = pd.to_datetime(end)  # type: pd.Timestamp

    # Assign the user's timezone, otherwise it gets very confusing as to what's being returned.
    pd_start = localize_datetime(session, pd_start)
    pd_end = localize_datetime(session, pd_end)

    if pd.isnull(pd_end):
        utc_now = pd.to_datetime(datetime.datetime.utcnow()).tz_localize('UTC')
        if not pd.isnull(pd_start):
            # noinspection PyTypeChecker
            pd_end = utc_now.tz_convert(get_user_timezone(session) if pd_start.tz is None else pd_start.tz)
            if pd_start > pd_end:
                # noinspection PyTypeChecker
                pd_end = pd_start + pd.Timedelta(hours=1)
        else:
            # noinspection PyTypeChecker
            pd_end = utc_now.tz_convert(get_user_timezone(session))

    if pd.isnull(pd_start):
        pd_start = pd_end - pd.Timedelta(hours=1)

    return pd_start, pd_end


# The list of units that come back from system_api.get_supported_units() is a curated set of compound units that are
# meant to be human-friendly in Formula tool help. But it's the only thing available in the API to figure out what
# base units are supported, so we have to split the compound unit string into its components. I.e., S/cm³ gets split
# into S and cm so that we can add those two "base" units to the set we use to determine validity.
UNITS_SPLIT_REGEX = r'[/*²³·]'


def is_valid_unit(session: Session, unit):
    """
    Returns True if the supplied unit will be recognized by the Seeq calculation engine. This can be an important
    function to use if you are attempting to supply a "Value Unit Of Measure" property on a Signal or a "Unit Of
    Measure" property on a Scalar.
    :param session: The login session (necessary to execute this call)
    :param unit: The unit of measure for which to assess validity
    :return: True if unit is valid, False if not
    """
    if not session.supported_units:
        system_api = SystemApi(session.client)
        support_units_output = system_api.get_supported_units()  # type: SupportedUnitsOutputV1

        session.supported_units = set()
        for supported_unit_family in support_units_output.units.values():
            for supported_unit in supported_unit_family:
                unit_parts = re.split(UNITS_SPLIT_REGEX, supported_unit)
                session.supported_units.update([u for u in unit_parts if len(u) > 0])

    unit_parts = re.split(UNITS_SPLIT_REGEX, unit)
    for unit_part in [u for u in unit_parts if len(u) > 0]:
        if unit_part not in session.supported_units:
            return False

    return True


def pull_image(session: Session, url):
    return requests.get(url, headers={
        "Accept": "application/vnd.seeq.v1+json",
        "x-sq-auth": session.client.auth_token
    }, verify=session.https_verify_ssl).content


def get_api(session: Session, api_class):
    validate_login(session, Status(quiet=True))
    return api_class(session.client)


def __getattr__(name):
    if name == 'client':
        # Before ~ March 2020, users commonly accessed spy._login.client when they wanted to leverage the SPy login
        # in their use of SDK functions.
        util.deprecation_warning("Use of spy._login.client deprecated, use spy.client instead")
        return spy.client
    else:
        # fallback to default module attribute access
        raise AttributeError



import pandas as pd
import pytest

from seeq import spy
from seeq.sdk import *
from seeq.spy import addons, Session
from seeq.spy.addons.tests import test_common as addons_test_common
from seeq.spy.tests import test_common
from seeq.spy.tests.test_common import Sessions


def setup_module():
    test_common.initialize_sessions()
    addons_test_common.enable_addon_tools(test_common.get_session(Sessions.admin), True)


# noinspection HttpUrlsUsage
def _create_test_tools(session: Session, suffix_=''):
    my_tools = [{"Name": f'test tool 1_{suffix_}',
                 "Description": f"test tool 1{suffix_}",
                 "Target URL": "http://www.google.com",
                 "Icon": "fa fa-icon",
                 "Link Type": "tab",
                 "Users": [session.user.username]},
                {"Name": f'test tool 2_{suffix_}',
                 "Description": f"test tool 2{suffix_}",
                 "Target URL": "http://www.seeq.com",
                 "Icon": "fa fa-icon",
                 "Users": [session.user.username]
                 }]
    # searching for tools doesn't require admin access but creating tools does
    tools = addons.install(my_tools, session=test_common.get_session(Sessions.admin))
    return tools


def _uninstall_test_tools(ids):
    for idd in ids:
        system_api = SystemApi(test_common.get_session(Sessions.admin).client)
        system_api.delete_add_on_tool(id=idd)


@pytest.mark.system
def test_addon_support():
    addons_test_common.test_addon_support(spy.session, 'search', dict())


@pytest.mark.system
def test_search_df_metadata():
    query = {'Name': '*'}
    search_results = addons.search(query)

    assert isinstance(search_results.spy.status.df, pd.DataFrame)
    assert search_results.spy.kwargs['query'] == query


@pytest.mark.system
def test_search_with_wildcard_plus_another_prop():
    unique_name_suffix = str(pd.to_datetime("today"))
    session = test_common.get_session(Sessions.nonadmin)
    df_tools = _create_test_tools(session, suffix_=unique_name_suffix)

    query = {'Name': '*', "Description": f"2{unique_name_suffix}"}
    search_results = addons.search(query, session=session)

    assert len(search_results) == 1
    # clean up
    _uninstall_test_tools(df_tools['ID'].values)


@pytest.mark.system
def test_search_with_wildcard_plus_id():
    unique_name_suffix = str(pd.to_datetime("today"))
    session = test_common.get_session(Sessions.nonadmin)
    df_tools = _create_test_tools(session, suffix_=unique_name_suffix)

    query = {'Name': '*', "ID": df_tools['ID'].values[0]}
    search_results = addons.search(query, session=session)

    assert len(search_results) == 1

    # clean up
    _uninstall_test_tools(df_tools['ID'].values)


@pytest.mark.system
def test_search_by_id():
    unique_name_suffix = str(pd.to_datetime("today"))
    session = test_common.get_session(Sessions.nonadmin)
    df_tools = _create_test_tools(session, suffix_=unique_name_suffix)
    idd = df_tools['ID'][0]
    search_results = addons.search({"ID": idd}, session=session)
    assert len(search_results) == 1
    # clean up
    _uninstall_test_tools(df_tools['ID'].values)


@pytest.mark.system
def test_search_with_df():
    unique_name_suffix = str(pd.to_datetime("today"))
    session = test_common.get_session(Sessions.nonadmin)
    df_tools = _create_test_tools(session, suffix_=unique_name_suffix)
    my_items = pd.DataFrame(
        {'Name': [f'test tool 1_{unique_name_suffix}', f'test tool 2_{unique_name_suffix}'],
         'Link Type': 'window'})
    search_results = addons.search(my_items, session=session)
    assert len(search_results) == 1
    assert search_results['Link Type'][0] == 'window'

    # clean up
    _uninstall_test_tools(df_tools['ID'].values)


@pytest.mark.system
def test_search_with_multiple_props():
    unique_name_suffix = str(pd.to_datetime("today"))
    session = test_common.get_session(Sessions.nonadmin)
    df_tools = _create_test_tools(session, suffix_=unique_name_suffix)

    search = addons.search({"Name": "test tool", "Description": "test tool"}, session=session)
    assert len(search) >= 2

    search_results_no_match = addons.search(
        pd.DataFrame([{"Name": f'test tool', "Description": "test tool"}]))
    search_results_match = addons.search(pd.DataFrame([{"Name": f'test tool 1_{unique_name_suffix}',
                                                        "Description": f"test tool 1{unique_name_suffix}"},
                                                       {"Name": f'test tool 2_{unique_name_suffix}',
                                                        "Description": f"test tool 2{unique_name_suffix}"}]),
                                         session=session)
    assert len(search_results_match) - len(search_results_no_match) == 2

    # clean up
    _uninstall_test_tools(df_tools['ID'].values)



import os
import tempfile

import pytest

from seeq.base import system
from seeq.spy.docs import _copy


@pytest.mark.unit
def test_copy():
    with tempfile.TemporaryDirectory() as temp_folder:
        long_folder_name = os.path.join(temp_folder, 'long_' * 10)
        _copy.copy(long_folder_name)

        assert os.path.exists(long_folder_name)

        for file_name in _copy.ADVANCED_ONLY_FILES:
            assert not os.path.exists(os.path.join(long_folder_name, file_name))

        with pytest.raises(RuntimeError):
            _copy.copy(long_folder_name)

        assert os.path.exists(long_folder_name)

        _copy.copy(long_folder_name, overwrite=True, advanced=True)

        assert os.path.exists(long_folder_name)
        assert os.path.exists(os.path.join(long_folder_name, 'spy.workbooks.ipynb'))
        assert os.path.exists(os.path.join(long_folder_name, 'Asset Trees 3 - Report and Dashboard Templates.ipynb'))

        system.removetree(long_folder_name)



import os
from pathlib import Path

import mock
import pytest

from seeq.base import system
from seeq.spy.jobs import _common as _jobs_common
from seeq.spy.jobs import _execute

execution_notebook_file = "data-lab/jupyter/scheduling/ExecutionNotebook.ipynb"
test_notebook_path = "sdk/pypi/seeq/spy/docs/Documentation"
test_notebook_file = "spy.pull.ipynb"
label = 'run-in-executor-label'
test_sched_index = '0'
job_key = '2ff2df94-a532-4f20-8864-959813ab4c17_7726F74691C5'

notebook_file_scheduled = system.cleanse_path(os.path.join(system.get_repo_root_dir(), test_notebook_path,
                                                           test_notebook_file))
scheduled_file_filename: str = Path(notebook_file_scheduled).name
scheduled_file_folder: Path = Path(notebook_file_scheduled).parent


def setup_module():
    # assume all the calls return successfully
    execution_notebook = system.cleanse_path(os.path.join(system.get_repo_root_dir(), execution_notebook_file))
    _jobs_common.get_execution_notebook = mock.Mock(return_value=execution_notebook)

    setup_run_in_executor()


def teardown_module():
    # reset values for other tests
    cleanup()


@pytest.mark.unit
def test_get_notebook_file_scheduled():
    assert _execute.get_notebook_file_scheduled() == notebook_file_scheduled


@pytest.mark.unit
def test_get_label():
    assert _execute.get_label() == label


@pytest.mark.unit
def test_get_index():
    assert _execute.get_index() == test_sched_index


@pytest.mark.unit
def test_get_label():
    assert _execute.get_job_key() == job_key


@pytest.mark.unit
def test_job_result_filename():
    _job_result_filename = _execute._compose_filename(test_sched_index, label, scheduled_file_filename,
                                                      scheduled_file_folder)

    expected_filename = os.path.join(system.get_repo_root_dir(), test_notebook_path, _jobs_common.get_results_folder(),
                                     f'spy.pull.executor.{test_sched_index}.{label}.html')
    assert system.cleanse_path(_job_result_filename) == system.cleanse_path(expected_filename)


@pytest.mark.unit
def test_job_temporary_filename():
    _job_result_filename = _execute._compose_filename(test_sched_index, label, scheduled_file_filename,
                                                      scheduled_file_folder, extension='.ipynb', result=False)

    expected_filename = os.path.join(system.get_repo_root_dir(), test_notebook_path, scheduled_file_folder,
                                     f'.spy.pull.executor.{test_sched_index}.{label}.ipynb')
    assert system.cleanse_path(_job_result_filename) == system.cleanse_path(expected_filename)


def setup_run_in_executor():
    os.environ['SEEQ_SDL_FILE_PATH'] = notebook_file_scheduled
    os.environ['SEEQ_SDL_LABEL'] = label
    os.environ['SEEQ_SDL_SCHEDULE_INDEX'] = test_sched_index
    os.environ['SEEQ_SDL_JOB_KEY'] = job_key


def cleanup():
    del os.environ['SEEQ_SDL_FILE_PATH']
    del os.environ['SEEQ_SDL_LABEL']
    del os.environ['SEEQ_SDL_SCHEDULE_INDEX']
    del os.environ['SEEQ_SDL_JOB_KEY']



import os

import mock
import pandas as pd
import pytest
import requests

from seeq import spy
from seeq.spy import Session
from seeq.spy._config import Setting
from seeq.spy.jobs import _schedule
from seeq.spy.jobs.tests import test_schedule_unit
from seeq.spy.tests import test_common

seeq_url = 'http://localhost:34216'


def setup_module():
    test_common.initialize_sessions()
    test_schedule_unit.setup_mocks()


@pytest.mark.system
def test_schedule_in_datalab():
    setup_run_in_datalab()

    test_jobs_df = pd.DataFrame({'Schedule': ['0 */2 1 * * ? *', '0 0 2 * * ? *', '0 42 03 22 1 ? 2121']})
    test_status = spy.Status()
    schedule_result = _schedule.schedule_df(spy.session, jobs_df=test_jobs_df, status=test_status)
    assert test_status.message.startswith("Scheduled")
    assert 'notebook.ipynb' in test_status.message
    assert 'notebook.pkl' in test_status.message
    assert len(schedule_result.index) == 3

    # dataframe without the schedule column, but first column containing the schedules
    datalab_notebook_url = f'{seeq_url}/data-lab/8A54CD8B-B47A-42DA-B8CC-38AD4204C865/notebooks/1.ipynb'
    test_jobs_df = pd.DataFrame({'Other Name': ['0 */2 1 * * ? *', '0 0 2 * * ? *']})
    test_status = spy.Status()
    schedule_result = _schedule.schedule_df(spy.session, jobs_df=test_jobs_df,
                                            datalab_notebook_url=datalab_notebook_url,
                                            status=test_status)
    assert test_status.message.startswith("Scheduled")
    assert '1.ipynb' in test_status.message
    assert '1.pkl' in test_status.message
    assert len(schedule_result.index) == 2

    # schedule with a label
    label = 'This is a label!'
    schedule_result = _schedule.schedule_df(spy.session, jobs_df=test_jobs_df,
                                            datalab_notebook_url=datalab_notebook_url,
                                            label=label, status=test_status)
    assert test_status.message.startswith("Scheduled")
    assert '1.ipynb' in test_status.message
    assert f' with label <strong>{label}</strong> ' in test_status.message
    assert f'1.with.label.{label}.pkl' in test_status.message
    assert len(schedule_result.index) == 2

    # dataframe without the schedule column
    test_jobs_df = pd.DataFrame({'Some Name': ['abc'], 'Other Name': ['abc']})
    with pytest.raises(ValueError, match='Could not interpret "abc" as a schedule'):
        _schedule.schedule_df(spy.session, jobs_df=test_jobs_df, status=test_status)


@pytest.mark.system
def test_schedule_in_executor():
    setup_run_in_executor()

    test_jobs_df = pd.DataFrame({'Schedule': ['0 */5 * * * ?'], 'Param': ['val1']})
    test_status = spy.Status()
    schedule_result = _schedule.schedule_df(spy.session, jobs_df=test_jobs_df, status=test_status)
    assert test_status.message.startswith("Scheduled")
    assert 'test.ipynb' in test_status.message
    assert 'test.pkl' in test_status.message
    assert len(schedule_result.index) == 1


@pytest.mark.system
def test_schedule_outside_datalab():
    setup_run_outside_datalab()
    test_common.log_out_default_user()

    test_jobs_df = pd.DataFrame({'Schedule': ['0 0 2 * * ? *', '0 42 03 22 1 ? 2121']})
    test_status = spy.Status()
    with pytest.raises(RuntimeError) as err1:
        _schedule.schedule_df(spy.session, jobs_df=test_jobs_df, status=test_status)
    assert "Not logged in" in str(err1.value)

    test_common.log_in_default_user()

    # no datalab_notebook_url provided
    with pytest.raises(RuntimeError) as err2:
        _schedule.schedule_df(spy.session, jobs_df=test_jobs_df, status=test_status)
    assert "Provide a Seeq Data Lab Notebook URL" in str(err2.value)

    # noinspection HttpUrlsUsage
    datalab_notebook_url = f'http://remote.com/data-lab/8A54CD8B-B47A-42DA-B8CC-38AD4204C865/notebooks/1.ipynb'
    schedule_result = _schedule.schedule_df(spy.session, jobs_df=test_jobs_df,
                                            datalab_notebook_url=datalab_notebook_url,
                                            status=test_status)
    assert test_status.message.startswith("Scheduled")
    assert '1.ipynb' in test_status.message
    assert '1.pkl' in test_status.message
    assert len(schedule_result.index) == 2


@pytest.mark.system
def test_unschedule_in_datalab():
    setup_run_in_datalab()

    test_status = spy.Status()
    schedule_result = _schedule.schedule_df(spy.session, status=test_status)
    assert test_status.message.startswith("Unscheduled")
    assert 'notebook.ipynb' in test_status.message
    pd.testing.assert_frame_equal(schedule_result, pd.DataFrame())

    datalab_notebook_url = f'{seeq_url}/data-lab/8A54CD8B-B47A-42DA-B8CC-38AD4204C865/notebooks/1.ipynb'
    schedule_result = _schedule.schedule_df(spy.session, datalab_notebook_url=datalab_notebook_url, status=test_status)
    assert test_status.message.startswith("Unscheduled")
    assert '1.ipynb' in test_status.message
    pd.testing.assert_frame_equal(schedule_result, pd.DataFrame())

    label = 'Labeled!'
    schedule_result = _schedule.schedule_df(spy.session, jobs_df=pd.DataFrame(), label=label, status=test_status)
    assert test_status.message.startswith("Unscheduled")
    assert 'notebook.ipynb' in test_status.message
    assert label in test_status.message
    pd.testing.assert_frame_equal(schedule_result, pd.DataFrame())

    label = '*'
    schedule_result = _schedule.schedule_df(spy.session, jobs_df=pd.DataFrame(), label=label, status=test_status)
    assert test_status.message.startswith("Unscheduled")
    assert 'notebook.ipynb' in test_status.message
    assert 'for all labels' in test_status.message
    pd.testing.assert_frame_equal(schedule_result, pd.DataFrame())


@pytest.mark.system
def test_unschedule_outside_datalab():
    setup_run_outside_datalab()
    test_common.log_out_default_user()

    test_status = spy.Status()

    with pytest.raises(RuntimeError) as err:
        _schedule.schedule_df(spy.session, status=test_status)
    assert "Not logged in" in str(err.value)

    test_common.log_in_default_user()

    # should provide a datalab_notebook_url
    with pytest.raises(RuntimeError):
        _schedule.schedule_df(spy.session, status=test_status)

    # noinspection HttpUrlsUsage
    datalab_notebook_url = f'http://remote.com/data-lab/8A54CD8B-B47A-42DA-B8CC-38AD4204C865/notebooks/1.ipynb'
    schedule_result = _schedule.schedule_df(spy.session, datalab_notebook_url=datalab_notebook_url, status=test_status)
    assert test_status.message.startswith("Unscheduled")
    assert '1.ipynb' in test_status.message
    pd.testing.assert_frame_equal(schedule_result, pd.DataFrame())

    schedule_result = _schedule.schedule_df(spy.session, jobs_df=pd.DataFrame(),
                                            datalab_notebook_url=datalab_notebook_url,
                                            status=test_status)
    assert test_status.message.startswith("Unscheduled")
    assert '1.ipynb' in test_status.message
    pd.testing.assert_frame_equal(schedule_result, pd.DataFrame())


@pytest.mark.system
def test_validate_and_get_next_trigger():
    validate_result = _schedule.validate_and_get_next_trigger(spy.session, ['0 */5 * * * ? 2069'])
    assert '2069-01-01 00:00:00 UTC' == validate_result['0 */5 * * * ? 2069']

    with pytest.raises(RuntimeError) as err1:
        _schedule.validate_and_get_next_trigger(spy.session, ['0 */5 * * * ? 2001', '0 */5 * * * *'])
    assert "schedules are invalid" in str(err1.value)
    assert "0 */5 * * * ? 2001" in str(err1.value)
    assert "No future trigger" in str(err1.value)

    with pytest.raises(RuntimeError) as err2:
        _schedule.validate_and_get_next_trigger(spy.session, ['* */2 * * * '])
    assert "Unexpected end of expression" in str(err2.value)

    with pytest.raises(RuntimeError) as err3:
        _schedule.validate_and_get_next_trigger(spy.session, ['* abc * * * ? *'])
    assert "Illegal characters for this position" in str(err3.value)


@pytest.mark.system
def test_retrieve_notebook_path_in_datalab():
    setup_run_in_datalab()
    data_lab_url, project_id, file_path = _schedule.retrieve_notebook_path(spy.session)
    assert project_id == '8A54CD8B-B47A-42DA-B8CC-38AD4204C863'
    assert file_path == 'notebook.ipynb'

    datalab_notebook_url = f'{seeq_url}/data-lab/8A54CD8B-B47A-42DA-B8CC-38AD4204C865/notebooks/1.ipynb'
    data_lab_url, project_id, file_path = _schedule.retrieve_notebook_path(spy.session, datalab_notebook_url)
    assert project_id == '8A54CD8B-B47A-42DA-B8CC-38AD4204C865'
    assert file_path == '1.ipynb'


@pytest.mark.system
def test_retrieve_notebook_path_in_executor():
    setup_run_in_executor()
    data_lab_url, project_id, file_path = _schedule.retrieve_notebook_path(spy.session)
    assert project_id == '8A54CD8B-B47A-42DA-B8CC-38AD4204C863'
    assert file_path == 'test.ipynb'

    datalab_notebook_url = f'{seeq_url}/data-lab/8A54CD8B-B47A-42DA-B8CC-38AD4204C866/notebooks/2.ipynb'
    data_lab_url, project_id, file_path = _schedule.retrieve_notebook_path(spy.session, datalab_notebook_url)
    assert project_id == '8A54CD8B-B47A-42DA-B8CC-38AD4204C866'
    assert file_path == '2.ipynb'


@pytest.mark.system
def test_retrieve_notebook_path_outside_sdl():
    setup_run_outside_datalab()
    with pytest.raises(RuntimeError) as err:
        _schedule.retrieve_notebook_path(spy.session)
    assert "Provide a Seeq Data Lab Notebook URL" in str(err.value)

    datalab_notebook_url = f'{seeq_url}/data-lab/8A54CD8B-B47A-42DA-B8CC-38AD4204C867/notebooks/3.ipynb'
    data_lab_url, project_id, file_path = _schedule.retrieve_notebook_path(spy.session, datalab_notebook_url)
    assert project_id == '8A54CD8B-B47A-42DA-B8CC-38AD4204C867'
    assert file_path == '3.ipynb'


@pytest.mark.system
def test_verify_not_actually_existing_and_accessible():
    setup_run_outside_datalab()
    # noinspection HttpUrlsUsage
    datalab_notebook_url = f'http://nonce.com/data-lab/8A54CD8B-B47A-42DA-B8CC-38AD4204C867/notebooks/3.ipynb'
    # noinspection HttpUrlsUsage
    contents_request_url = f'http://nonce.com/data-lab/8A54CD8B-B47A-42DA-B8CC-38AD4204C867/api/contents/3.ipynb'
    mock_resp = requests.Response()
    mock_resp.status_code = 404
    mock_requests_get = mock.Mock(return_value=mock_resp)
    session = Session()
    session.client = mock.MagicMock()
    session.client.auth_token = 'some token'
    with mock.patch('seeq.spy.jobs._common.requests_get', mock_requests_get):
        with pytest.raises(RuntimeError) as err:
            _schedule._verify_existing_and_accessible(session, datalab_notebook_url)
        assert 'Notebook not found for URL' in str(err.value)
        mock_requests_get.assert_called_once_with(session,
                                                  contents_request_url,
                                                  params={'type': 'file', 'content': 0},
                                                  cookies={'sq-auth': 'some token'})


def setup_run_outside_datalab():
    os.environ.pop('SEEQ_SDL_CONTAINER_IS_DATALAB', None)
    os.environ.pop('SEEQ_SDL_CONTAINER_IS_EXECUTOR', None)
    os.environ.pop('SEEQ_PROJECT_UUID', None)
    os.environ.pop('SEEQ_SDL_FILE_PATH', None)
    os.environ.pop('SEEQ_SDL_LABEL', None)
    os.environ.pop('SEEQ_SDL_SCHEDULE_INDEX', None)


def setup_run_in_datalab():
    os.environ['SEEQ_SDL_CONTAINER_IS_DATALAB'] = 'true'
    os.environ['SEEQ_SDL_CONTAINER_IS_EXECUTOR'] = ''
    os.environ['SEEQ_PROJECT_UUID'] = '8A54CD8B-B47A-42DA-B8CC-38AD4204C863'
    Setting.set_seeq_url(seeq_url)
    Setting.SEEQ_PROJECT_UUID.set('8A54CD8B-B47A-42DA-B8CC-38AD4204C863')


def setup_run_in_executor():
    os.environ['SEEQ_SDL_CONTAINER_IS_DATALAB'] = ''
    os.environ['SEEQ_SDL_CONTAINER_IS_EXECUTOR'] = 'true'
    os.environ['SEEQ_PROJECT_UUID'] = '8A54CD8B-B47A-42DA-B8CC-38AD4204C864'
    os.environ['SEEQ_SDL_FILE_PATH'] = 'test.ipynb'
    os.environ['SEEQ_SDL_LABEL'] = 'run-in-executor-label'
    os.environ['SEEQ_SDL_SCHEDULE_INDEX'] = '0'
    Setting.set_seeq_url(seeq_url)
    Setting.SEEQ_PROJECT_UUID.set('8A54CD8B-B47A-42DA-B8CC-38AD4204C863')



import base64
import json
import os
import pathlib
import pickle
import re
import types
from datetime import datetime, timedelta
from typing import Optional
from urllib.parse import unquote

import cron_descriptor
import ipykernel
import pandas as pd
import pytz as tz
from recurrent import RecurringEvent

from seeq.sdk import *
from seeq.spy import _common, _config, _login
from seeq.spy._errors import *
from seeq.spy._session import Session
from seeq.spy._status import Status
from seeq.spy.jobs import _common as _jobs_common


def schedule(schedule_spec: str, datalab_notebook_url: Optional[str] = None, label: Optional[str] = None,
             user: Optional[str] = None, suspend: bool = False, quiet: bool = False,
             status: Optional[Status] = None, session: Optional[Session] = None):
    """
    Schedules the automatic execution of a Seeq Data Lab notebook.

    The current notebook is scheduled for execution unless datalab_notebook_url
    is supplied. Scheduling can be done on behalf of another user by a user with
    admin privileges.

    Successive calls to 'schedule()' for the same notebook and label but with
    different schedules will replace the previous schedule for the notebook-
    label combination.

    Removing the scheduling is accomplished via unschedule().

    A copy of the jobs DataFrame is automatically stored to a _Job DataFrames
    folder adjacent to the Notebook for which the job is scheduled.

    Parameters
    ----------
    schedule_spec : str
        A string that represents the frequency with which the notebook should
        execute.

        Examples: 'every 15 minutes'
                  'every tuesday and friday at 6am'
                  'every fifth of the month'

        The timezone used for scheduling is the one specified in the logged-in
        user's profile.

        You can also use Quartz Cron syntax. Use the following site to
        construct it:
        https://www.freeformatter.com/cron-expression-generator-quartz.html

    datalab_notebook_url : str, default None
        A datalab notebook URL. If the value is not specified the currently
        running notebook URL is used.

    label : str, default None
        A string used to enable scheduling of the Notebook by different users
        or from different Analysis Pages.  Labels may contain letters, numbers,
        spaces, and the following special characters: !@#$^&-_()[]{}

    user : str, default None
        Determines the user on whose behalf the notebook is executed. If the
        value is not specified the currently logged in user is used. The can be
        specified as username or a user's Seeq ID.

    suspend : bool, default False
        If True, unschedules all jobs for the specified notebook. This is used
        in scenarios where you wish to work with a notebook interactively and
        temporarily "pause" job execution. Remove the argument (or change it
        to False) when you are ready to resume job execution.

    quiet : bool, default False
        If True, suppresses progress output. Note that when status is
        provided, the quiet setting of the Status object that is passed
        in takes precedent.

    status : spy.Status, optional
        If specified, the supplied Status object will be updated as the command
        progresses. It gets filled in with the same information you would see
        in Jupyter in the blue/green/red table below your code while the
        command is executed. The table itself is accessible as a DataFrame via
        the status.df property.

    session : spy.Session, optional
        If supplied, the Session object (and its Options) will be used to
        store the login session state. This is useful to log in to different
        Seeq servers at the same time or with different credentials.

    Returns
    -------
    pd.DataFrame
        The jobs_df with an appended column containing a description of the
        schedule
    """
    _common.validate_argument_types([
        (schedule_spec, 'schedule_string', str),
        (datalab_notebook_url, 'datalab_notebook_url', str),
        (label, 'label', str),
        (user, 'user', str),
        (suspend, 'suspend', bool),
        (quiet, 'quiet', bool),
        (status, 'status', Status),
        (session, 'session', Session)
    ])

    status = Status.validate(status, quiet)
    session = Session.validate(session)

    try:
        return schedule_df(session, pd.DataFrame([{'Schedule': schedule_spec}]) if schedule_spec else None,
                           datalab_notebook_url=datalab_notebook_url, label=label, user=user, suspend=suspend,
                           status=status)
    except SchedulePostingError:
        # See _push.push() for why we swallow this error in the executor
        if not _jobs_common.running_in_executor():
            raise


def unschedule(datalab_notebook_url: Optional[str] = None, label: Optional[str] = None, quiet: bool = False,
               status: Optional[Status] = None, session: Optional[Session] = None):
    """
    Unschedules ALL jobs for a particular notebook and label.

    The current notebook is unscheduled unless datalab_notebook_url
    is supplied. Unscheduling can be done on behalf of another user by a user
    with admin privileges.

    Parameters
    ----------
    datalab_notebook_url : str, default None
        A datalab notebook URL. If the value is not specified the currently
        running notebook URL is used.

    label : str, default None
        A string used to enable scheduling of the Notebook by different users
        or from different Analysis Pages.  Labels may contain letters, numbers,
        spaces, and the following special characters:

        !@#$^&-_()[]{}

        A value of '*' will unschedule all jobs across all labels associated
        with the supplied notebook (or the current Notebook, if no
        datalab_notebook_url is supplied).

    quiet : bool, default False
        If True, suppresses progress output. Note that when status is
        provided, the quiet setting of the Status object that is passed
        in takes precedent.

    status : spy.Status, optional
        If specified, the supplied Status object will be updated as the command
        progresses. It gets filled in with the same information you would see
        in Jupyter in the blue/green/red table below your code while the
        command is executed. The table itself is accessible as a DataFrame via
        the status.df property.

    session : spy.Session, optional
        If supplied, the Session object (and its Options) will be used to
        store the login session state. This is useful to log in to different
        Seeq servers at the same time or with different credentials.
    """
    _common.validate_argument_types([
        (datalab_notebook_url, 'datalab_notebook_url', str),
        (label, 'label', str),
        (quiet, 'quiet', bool),
        (status, 'status', Status),
        (session, 'session', Session)
    ])

    status = Status.validate(status, quiet)
    session = Session.validate(session)
    schedule_df(session, jobs_df=None, datalab_notebook_url=datalab_notebook_url, label=label, status=status)


def schedule_df(session: Session, jobs_df=None, spread=None, datalab_notebook_url=None, label=None, user=None,
                suspend=False, status=None):
    input_args = _common.validate_argument_types([
        (jobs_df, 'jobs_df', pd.DataFrame),
        (datalab_notebook_url, 'datalab_notebook_url', str),
        (label, 'label', str),
        (user, 'user', str),
        (suspend, 'suspend', bool),
        (status, 'status', Status)
    ])

    if jobs_df is None:
        jobs_df = pd.DataFrame()

    _login.validate_login(session, status)

    _common.validate_unique_dataframe_index(jobs_df, 'jobs_df')

    indexed_cron_expressions = _get_cron_expression_list(jobs_df) if not suspend else []
    if spread:
        indexed_cron_expressions = _spread_over_period(indexed_cron_expressions, spread)
    next_trigger_map = validate_and_get_next_trigger(session, [cron for idx, cron in indexed_cron_expressions])
    datalab_base_url, project_id, file_path = retrieve_notebook_path(session, datalab_notebook_url)
    # Even though the following verify method immediately calls retrieve_notebook_path again, it is convenient to have
    # the passed URL for the error message, so we don't call it with the resulting tuple.  The verification is done
    # separately so that retrieve_notebook_path can be reused for cases where existence of the URL is not required,
    # and so that an appropriate error message will be provided before attempting the contents API if the URL is bad.
    if datalab_notebook_url:
        _verify_existing_and_accessible(session, datalab_notebook_url)

    try:
        user_identity = _login.find_user(session, user) if user is not None else None
        if len(indexed_cron_expressions) == 0:
            _call_unschedule_notebook_api(session, project_id, file_path, label)
        else:
            _call_schedule_notebook_api(session, indexed_cron_expressions, project_id, file_path, label, user_identity)
    except BaseException as e:
        status.exception(e)
        raise SchedulePostingError(e)

    with_label_text = f' with label <strong>{label}</strong> ' if label else ' '
    if indexed_cron_expressions:
        status.update(f'Scheduled the notebook <strong>{file_path}</strong>{with_label_text}successfully.\n'
                      f'Current context is <strong>{"JOB" if _jobs_common.running_in_executor() else "INTERACTIVE"}'
                      '</strong>.', Status.SUCCESS)
        try:
            pickle_path = _pickle_jobs_df(session, jobs_df, datalab_notebook_url, label)
            if pickle_path:
                status.update(f'{status.message}  The jobs DataFrame was stored to {pickle_path}')
        except SPyRuntimeError as rue:
            status.update(f'Jobs DataFrame could not be stored due to error: {rue}')

    elif label == '*':
        status.update(f'Unscheduled all jobs for notebook <strong>{file_path}'
                      f'</strong> for all labels successfully.  Job DataFrame pickle files in '
                      f'_Job DataFrames subfolder must be cleared or removed manually when '
                      f'unscheduling with a label of \'*\'', Status.SUCCESS)
    else:
        pickle_path = _pickle_jobs_df(session, jobs_df, datalab_notebook_url, label)
        unlabeled = " " if label else " unlabeled "
        status.update(f'Unscheduled all{unlabeled}jobs for notebook <strong>{file_path}'
                      f'</strong>{with_label_text}successfully.', Status.SUCCESS)
        if pickle_path:
            status.update(f'{status.message}  The Job DataFrame at {pickle_path} has been cleared.')

    if indexed_cron_expressions:
        cron_expressions = [cron for idx, cron in indexed_cron_expressions]
        schedule_result_df = pd.concat(
            [jobs_df,
             pd.Series(cron_expressions, index=jobs_df.index).rename('Scheduled').map(cron_descriptor.get_description)],
            axis=1
        )
        schedule_result_df = pd.concat(
            [schedule_result_df,
             pd.Series(cron_expressions, index=jobs_df.index).rename('Next Run').map(lambda ce: next_trigger_map[ce])],
            axis=1
        )
    else:
        schedule_result_df = pd.DataFrame()

    schedule_df_properties = types.SimpleNamespace(
        func='spy.schedule',
        kwargs=input_args,
        status=status)

    _common.put_properties_on_df(schedule_result_df, schedule_df_properties)

    status.df = schedule_result_df
    status.update()
    return schedule_result_df


def _verify_existing_and_accessible(session: Session, datalab_notebook_url):
    datalab_base_url, project_id, file_path = retrieve_notebook_path(session, datalab_notebook_url)
    error = SPyRuntimeError(f'Notebook not found for URL {datalab_notebook_url}.  Verify that the notebook exists '
                            f'and the user account has access to it')
    if is_referencing_this_project(datalab_notebook_url):
        if not path_exists(pathlib.PurePosixPath(_common.DATALAB_HOME, file_path)):
            raise error
    else:
        contents_api_url = f'{datalab_base_url}/{project_id}/api/contents'
        params = {'type': 'file', 'content': 0}
        cookies = {"sq-auth": session.client.auth_token}
        resp = _jobs_common.requests_get(session, f'{contents_api_url}/{file_path}', params=params, cookies=cookies)
        if resp.status_code != 200:
            raise error


def _pickle_jobs_df(session: Session, jobs_df, datalab_notebook_url=None, label=None):
    data_lab_url, project_id, file_path = retrieve_notebook_path(session, datalab_notebook_url)
    file_path_path = pathlib.PurePosixPath(file_path)
    path_to_parent = path_or_empty_string(file_path_path.parent)
    with_label_text = f'.with.label.{label}' if label else ''
    jobs_df_pickle = f'{file_path_path.stem}{with_label_text}.pkl'

    jobs_dfs_folder_path = pathlib.PurePosixPath(path_to_parent, _common.JOB_DATAFRAMES_FOLDER_NAME)
    cookies = {'sq-auth': session.client.auth_token}
    contents_api_url = f'{data_lab_url}/{project_id}/api/contents'

    if not get_or_create_folder(session, contents_api_url, str(jobs_dfs_folder_path)):
        raise SPyRuntimeError(f'Could not get or create {_common.JOB_DATAFRAMES_FOLDER_NAME} folder')

    put_pickle_path = pathlib.PurePosixPath(jobs_dfs_folder_path, jobs_df_pickle)

    if is_referencing_this_project(datalab_notebook_url):
        dump_pickle(jobs_df, put_pickle_path)
        return put_pickle_path
    else:
        content_encoded = base64.b64encode(pickle.dumps(jobs_df)).decode('utf-8')
        content_model = {
            'format': 'base64',
            'name': jobs_df_pickle,
            'path': str(put_pickle_path),
            'type': 'file',
            'content': content_encoded
        }
        resp = _jobs_common.requests_put(session,
                                         f'{contents_api_url}/{put_pickle_path}',
                                         data=json.dumps(content_model),
                                         cookies=cookies)
        if resp.status_code in [200, 201]:
            return put_pickle_path
        else:
            raise SPyRuntimeError(f'Could not retrieve job due to error calling Jupyter Contents API: '
                                  f'({resp.status_code}) {resp.reason}')


def get_or_create_folder(session: Session, contents_api_url: str, folder_path: str):
    if is_referencing_this_project(contents_api_url):
        absolute_path = pathlib.PurePosixPath(_common.DATALAB_HOME, folder_path)
        if not path_exists(absolute_path):
            make_dirs(absolute_path)
        return True
    else:
        cookies = {"sq-auth": session.client.auth_token}
        folder_path_as_path = pathlib.PurePosixPath(folder_path)
        folder_parent = path_or_empty_string(folder_path_as_path.parent)
        resp = _jobs_common.requests_get(session, f'{contents_api_url}/{folder_parent}', cookies=cookies)
        if folder_path_as_path.name in [item['name'] for item in resp.json()['content']]:
            return True
        else:
            new_folder_post_resp = _jobs_common.requests_post(session,
                                                              contents_api_url,
                                                              data=json.dumps({'type': 'directory'}),
                                                              cookies=cookies)
            new_folder_path = pathlib.PurePosixPath(new_folder_post_resp.json()['path'])
            rename_folder_resp = _jobs_common.requests_patch(session,
                                                             f'{contents_api_url}/{new_folder_path}',
                                                             data=json.dumps({'path': folder_path}),
                                                             cookies=cookies)
            return True if rename_folder_resp.status_code == 200 else False


def path_or_empty_string(path):
    return '' if str(path) in ['.', '/'] else str(path)


def pickle_df_to_path(df, path):
    df.to_pickle(path)


def path_exists(path):
    return os.path.exists(path)


def make_dirs(path):
    os.makedirs(path, exist_ok=True)


def dump_pickle(df, path):
    with open(pathlib.PurePosixPath(_common.DATALAB_HOME, path), mode='wb') as pickle_file:
        pickle.dump(df, pickle_file)


def validate_and_get_next_trigger(session: Session, cron_expression_list):
    jobs_api = JobsApi(session.client)
    validate_cron_input = ValidateCronListInputV1()
    validate_cron_input.timezone = _login.get_user_timezone(session)
    validate_cron_input.next_valid_time_after = datetime.now(tz.timezone(_login.get_user_timezone(session))) \
        .replace(microsecond=0)
    validate_cron_input.schedules = cron_expression_list
    validation_output = jobs_api.validate_cron(body=validate_cron_input)

    invalid_cron_expressions = []
    for cron_validation in validation_output.schedules:
        if cron_validation.error is not None:
            invalid_cron_expressions.append({'string': cron_validation.quartz_cron_expression,
                                             'error': cron_validation.error})
    if invalid_cron_expressions:
        raise SPyRuntimeError('The following schedules are invalid: ' + str(invalid_cron_expressions))
    return dict((s.quartz_cron_expression, s.next_run_time) for s in validation_output.schedules)


def _call_schedule_notebook_api(session: Session, cron_expressions, project_id, file_path, label, user_identity):
    projects_api = ProjectsApi(session.client)
    schedule_input = ScheduledNotebookInputV1()
    schedule_input.file_path = file_path
    schedule_input.schedules = [ScheduleInputV1(key=idx, cron_schedule=cron) for idx, cron in cron_expressions]
    schedule_input.timezone = _login.get_user_timezone(session)
    schedule_input.label = label
    if user_identity is not None:
        schedule_input.user_id = user_identity.id

    projects_api.schedule_notebook(id=project_id, body=schedule_input)


def _call_unschedule_notebook_api(session: Session, project_id, file_path, label):
    if label is None:
        ProjectsApi(session.client).unschedule_notebook(id=project_id, file_path=file_path)
    else:
        ProjectsApi(session.client).unschedule_notebook(id=project_id, file_path=file_path, label=label)


def retrieve_notebook_path(session: Session, datalab_notebook_url=None):
    if datalab_notebook_url is None:
        if _jobs_common.running_in_datalab():
            return _config.get_data_lab_orchestrator_url(), _config.get_data_lab_project_id(), \
                   get_notebook_path(session)
        if _jobs_common.running_in_executor():
            return _config.get_data_lab_orchestrator_url(), _config.get_data_lab_project_id(), \
                   os.environ['SEEQ_SDL_FILE_PATH']
        else:
            raise SPyRuntimeError('Provide a Seeq Data Lab Notebook URL for scheduling')
    else:
        data_lab_url, project_id, file_path = parse_data_lab_url_project_id_and_path(datalab_notebook_url)
        if not data_lab_url:
            if _jobs_common.running_in_datalab() or _jobs_common.running_in_executor():
                data_lab_url = _config.get_data_lab_orchestrator_url()
            else:
                raise SPyRuntimeError('Path for Seeq Data Lab Notebook URL must include protocol and host if not '
                                      'running in Data Lab')
        return data_lab_url, project_id, file_path


def is_referencing_this_project(datalab_notebook_url):
    if datalab_notebook_url is None:
        return True
    else:
        return datalab_notebook_url.lower().startswith(_config.get_data_lab_project_url().lower())


def parse_data_lab_url_project_id_and_path(notebook_url):
    matches = re.search(r'^(.*/data-lab)/([0-9A-F]{8}-[0-9A-F]{4}-[0-9A-F]{4}-[0-9A-F]{4}-[0-9A-F]{12})/[\w]+/(.*)',
                        notebook_url, re.IGNORECASE)
    if not matches:
        raise SPyValueError('URL is not a valid SDL notebook')

    data_lab_url = _config.get_data_lab_orchestrator_url() if matches.group(1) == '/data-lab' else matches.group(1)
    project_id = matches.group(2).upper()
    file_path = unquote(matches.group(3))
    return data_lab_url, project_id, file_path


def get_notebook_path(session: Session):
    kernel_id = re.search('kernel-(.*).json', ipykernel.connect.get_connection_file()).group(1)
    cookies = {
        'sq-auth': session.client.auth_token
    }
    url = f'{_config.get_data_lab_project_url()}/api/sessions'
    response = _jobs_common.requests_get(session, url, cookies=cookies)
    for nn in json.loads(response.text):
        if nn['kernel']['id'] == kernel_id:
            return nn['notebook']['path']


def _spread_over_period(cron_expressions, over_period):
    over_period_delta = _common.parse_str_time_to_timedelta(over_period)

    if len(cron_expressions) == 0:
        return []

    if over_period_delta.total_seconds() > (60 * 60 * 24):
        raise SPyValueError(f'over_period cannot be more than 24 hours')

    spacing = timedelta(seconds=(over_period_delta.total_seconds() / len(cron_expressions)))
    current_slot = timedelta()
    new_expressions = list()
    for idx, cron_expression in cron_expressions:
        parts = re.split(r'\s+', cron_expression)
        parts[0] = re.sub(r'(.+?)(/.+)?', rf'{int(current_slot.total_seconds() % 60)}\2', parts[0])
        if over_period_delta.total_seconds() > 60:
            parts[1] = re.sub(r'(.+?)(/.+)?', rf'{int((current_slot.total_seconds() / 60) % 60)}\2', parts[1])
        if over_period_delta.total_seconds() > (60 * 60):
            parts[2] = re.sub(r'(.+?)(/.+)?', rf'{int(current_slot.total_seconds() / (60 * 60))}\2', parts[2])
        new_expressions.append((idx, ' '.join(parts)))
        current_slot += spacing

    return new_expressions


def _get_cron_expression_list(jobs_df):
    if jobs_df is None or jobs_df.empty:
        return []

    if 'Schedule' not in jobs_df:
        schedule_column = jobs_df.columns[0]
    else:
        schedule_column = 'Schedule'

    return [(idx, parse_schedule_string(row[schedule_column])) for idx, row in jobs_df.iterrows()]


def parse_schedule_string(schedule_string):
    # noinspection PyBroadException
    try:
        # If cron_descriptor can parse it, then it's a valid cron schedule already
        cron_descriptor.get_description(schedule_string)
        return schedule_string
    except BaseException:
        pass

    return friendly_schedule_to_quartz_cron(schedule_string)


def friendly_schedule_to_quartz_cron(schedule_string):
    r = RecurringEvent()
    r.parse(schedule_string)

    rules = r.get_params()
    if len(rules) == 0 or 'freq' not in rules or 'interval' not in rules:
        raise SPyValueError(f'Could not interpret "{schedule_string}" as a schedule')
    days = {'SU': 1, 'MO': 2, 'TU': 3, 'WE': 4, 'TH': 5, 'FR': 6, 'SA': 7}
    if not r.is_recurring:
        raise SPyValueError(f'Could not interpret "{schedule_string}" as a recurring schedule')
    second = '0'
    minute = rules.get('byminute', '0')
    hour = rules.get('byhour', '0')
    day_of_month = rules.get('bymonthday', '?')
    month = rules.get('bymonth', '*')
    day_of_week = ','.join([str(days[byday]) for byday in rules['byday'].split(',')]) if 'byday' in rules else '?'

    if day_of_month == '?' and day_of_week == '?':
        day_of_month = '1' if rules['freq'] in ['monthly', 'yearly'] else '*'

    interval = rules['interval']
    if rules['freq'] == 'minutely':
        hour = '*'
        minute = f'*/{interval}'
    elif rules['freq'] == 'hourly':
        hour = f'*/{interval}'
    elif rules['freq'] == 'daily':
        day_of_month += f'/{interval}'
    elif rules['freq'] == 'weekly':
        pass
    elif rules['freq'] == 'monthly':
        month += f'/{interval}'
    elif rules['freq'] == 'secondly':
        second, minute, hour, day_of_month = convert_seconds(interval)

    cron = f'{second} {minute} {hour} {day_of_month} {month} {day_of_week}'

    return cron


def convert_seconds(seconds_val: str) -> list:
    """
    Convert seconds value to appropriate time value.
    If it is detected that the time value is partial it will raise an exception.
    """

    t = timedelta(seconds=int(seconds_val))
    time_values = {
        'seconds': t.seconds % 60,
        'days': t.days,
        'hours': t.seconds // 3600,
        'minutes': (t.seconds // 60) % 60
    }

    # check on partial values
    if len([val for val in time_values.values() if val]) != 1:
        raise SPyValueError(f'Seconds value {seconds_val} could not be used for scheduling".')

    value_templates = {'days': '0;0;0;*/{}', 'hours': '0;0;*/{};*', 'minutes': '0;*/{};*;*', 'seconds': '*/{};*;*;*'}
    for time_name, time_val in time_values.items():
        if time_val:
            return value_templates[time_name].format(time_val).split(';')



import pandas as pd
import pytest

from seeq.spy import addons
from seeq.spy.addons.tests import test_common as addons_test_common
from seeq.spy.tests import test_common
from seeq.spy.tests.test_common import Sessions


def setup_module():
    test_common.initialize_sessions()
    addons_test_common.enable_addon_tools(test_common.get_session(Sessions.admin), True)


@pytest.mark.system
def test_addon_support():
    admin_session = test_common.get_session(Sessions.admin)
    addons_test_common.test_addon_support(admin_session, 'uninstall', pd.DataFrame())


# noinspection HttpUrlsUsage
@pytest.mark.system
def test_uninstall_tools():
    admin_session = test_common.get_session(Sessions.admin)
    unique_name_suffix = str(pd.to_datetime("today"))
    my_tools = [{"Name": f'test tool 1_{unique_name_suffix}',
                 "Description": "test tool 1",
                 "Target URL": "http://www.google.com",
                 "Icon": "fa fa-icon",
                 "Link Type": "tab"},
                {"Name": f'test tool 2_{unique_name_suffix}',
                 "Description": "test tool 2",
                 "Target URL": "http://www.seeq.com",
                 "Icon": "fa fa-icon"}]
    df_tools = addons.install(my_tools, session=admin_session)
    search_results = addons.search(df_tools, session=admin_session)

    assert len(search_results) == 2
    addons.uninstall(search_results, session=admin_session)
    search_results = addons.search(df_tools, session=admin_session)

    assert search_results.empty



import copy
import re

from seeq import spy
from seeq.sdk import *
from seeq.spy import _common
from seeq.spy import _login
from seeq.spy._errors import *
from seeq.spy._session import Session
from seeq.spy.workbooks._item import Item


class Content(Item):
    """
    The SPy representation of a Seeq Content Item. Content can be created with a number of different sizing
    parameters defined in the display() function within an HTML template.

    If no size, shape, width, or height are specified in the display() function, size and shape will default to medium
    and rectangle, respectively. If a height and width are specified, they will take precedence over size and shape.
    A height must be specified with a width and vice-versa.

    ===================== ==================================================
    Input Column          Content Size Attribute
    ===================== ==================================================
    Size                  The size of the content. Can be 'small', 'medium',
                          or 'large'. Defaults to 'medium'
    Shape                 The content's shape. Can be 'strip', 'rectangle',
                          or 'square'. Defaults to 'rectangle'
    Height                The height of the content. Takes precedence over
                          the size/shape parameter, if specified
    Width                 The width of the content. Takes precedence over
                          the size/shape parameter, if specified
    Scale                 The content's desired scale. A value greater than
                          1 will increase the size of elements within the
                          screenshot. A value less than 1 will shrink the
                          size of elements within the screenshot
    selector              The content's css styling.  The
                          '.screenshotSizeToContent' style can be used to
                          trim whitespace content, which is useful for
                          tables, which have an arbitrary shape.
                          Defaults to None.
    """
    CONTENT_SHAPE = {'strip': {'width': 16, 'height': 4},
                     'rectangle': {'width': 16, 'height': 9},
                     'square': {'width': 15, 'height': 15}}

    CONTENT_SIZE = {'small': 350, 'medium': 700, 'large': 1050}

    def __init__(self, definition, report=None):
        super().__init__(definition)
        self.report = report

    @property
    def name(self):
        # Since we don't have a way of uniquely identifying a piece of Content that results from a display()
        # declaration in an HTML report template, we instead hash some state for Content items and use that
        # as the name. Then we can use that name to match up content from an HTML template with an in-memory
        # object.
        return self.definition_hash

    @property
    def definition_hash(self):
        definition_copy = copy.deepcopy(self.definition)
        for key in ['ID', 'Name']:
            if key in definition_copy:
                del definition_copy[key]
        return Item.digest_hash(definition_copy)

    def push(self, session: Session, item_map: dict, existing_contents: dict):
        session = Session.validate(session)
        content_input = self._create_content_input(item_map)

        content_to_push = self.apply_map(item_map)
        existing_content = existing_contents.get(content_to_push.id)

        if not existing_content:
            existing_content = existing_contents.get(content_to_push.name)

        content_api = ContentApi(session.client)
        if not existing_content:
            content_output = content_api.create_content(body=content_input)  # type: ContentOutputV1
        else:
            content_output = content_api.update_content(
                id=existing_content.id, body=content_input)  # type: ContentOutputV1

        item_map[self.id.upper()] = content_output.id.upper()

        return content_output

    def _create_content_input(self, item_map=None):
        self._validate_fields_before_push()

        date_range_id = None
        if _common.present(self.definition, 'Date Range ID'):
            if self.definition['Date Range ID'].upper() not in item_map:
                raise SPyDependencyNotFound(
                    f'Date Range {self.definition["Date Range ID"].upper()} for {self} not found')

            date_range_id = item_map[self.definition['Date Range ID'].upper()]

        asset_selection_id = None
        if _common.present(self.definition, 'Asset Selection ID'):
            if self.definition['Asset Selection ID'].upper() not in item_map:
                raise SPyDependencyNotFound(
                    f'Asset Selection {self.definition["Asset Selection ID"].upper()} for {self} not found')

            asset_selection_id = item_map[self.definition['Asset Selection ID'].upper()]

        if self.definition['Worksheet ID'] not in item_map:
            raise SPyDependencyNotFound(f'Worksheet {self.definition["Worksheet ID"]} not found')

        if self.definition['Workstep ID'] not in item_map:
            raise SPyDependencyNotFound(f'Workstep {self.definition["Workstep ID"]} not found')

        # Report has definitely already been pushed by this time
        report_id = item_map[self.report.id.upper()]

        scale = self.definition['Scale'] if 'Scale' in self.definition else None

        summary_type = self.definition['Summary Type'] if 'Summary Type' in self.definition else None
        summary_value = self.definition['Summary Value'] if 'Summary Value' in self.definition else None

        return ContentInputV1(name=self.definition['Name'],
                              width=self.definition['Width'],
                              height=self.definition['Height'],
                              worksheet_id=item_map[self.definition['Worksheet ID']],
                              workstep_id=item_map[self.definition['Workstep ID']],
                              date_range_id=date_range_id,
                              asset_selection_id=asset_selection_id,
                              report_id=report_id,
                              scale=scale,
                              summary_type=summary_type,
                              summary_value=summary_value,
                              selector=self.definition['selector'],
                              archived=_common.get(self.definition, 'Archived', False))

    def _validate_fields_before_push(self):
        if 'Name' not in self.definition:
            raise SPyValueError('Unable to push Content: missing "Name" field')
        if 'Width' not in self.definition:
            raise SPyValueError('Unable to push Content: missing "Width" field')
        if 'Height' not in self.definition:
            raise SPyValueError('Unable to push Content: missing "Height" field')
        if 'Worksheet ID' not in self.definition:
            raise SPyValueError('Unable to push Content: missing "Worksheet ID" field')
        if 'Workstep ID' not in self.definition:
            raise SPyValueError('Unable to push Content: missing "Workstep ID" field')
        if 'selector' not in self.definition:
            self.definition['selector'] = ''

    @staticmethod
    def pull(item_id, *, allowed_types=None, report=None, session: Session = None):
        session = Session.validate(session)
        content_api = ContentApi(session.client)
        content_output = content_api.get_content(id=item_id)  # type: ContentOutputV1

        new_content_definition = {'Name': content_output.name,
                                  'ID': content_output.id,
                                  'Width': content_output.width,
                                  'Height': content_output.height,
                                  'Worksheet ID': content_output.source_worksheet,
                                  'Workstep ID': content_output.source_workstep,
                                  'Workbook ID': content_output.source_workbook,
                                  'selector': content_output.selector,
                                  'Summary Type': content_output.summary_type,
                                  'Summary Value': content_output.summary_value,
                                  'Scale': content_output.scale}

        if content_output.date_range is not None:
            new_content_definition['Date Range ID'] = content_output.date_range.id

        if content_output.asset_selection is not None:
            new_content_definition['Asset Selection ID'] = content_output.asset_selection.id

        return Content(new_content_definition, report)

    def construct_data_id(self, label):
        return self._construct_data_id(label)

    @property
    def date_range(self):
        if 'Date Range ID' not in self.definition:
            return None

        return self.report.date_ranges[self.definition['Date Range ID']]

    @date_range.setter
    def date_range(self, value):
        if value is None:
            self.definition['Date Range ID'] = None
        else:
            self.definition['Date Range ID'] = value['ID']

    @property
    def asset_selection(self):
        if 'Asset Selection ID' not in self.definition:
            return None

        return self.report.asset_selections[self.definition['Asset Selection ID']]

    @asset_selection.setter
    def asset_selection(self, value):
        if value is None:
            self.definition['Asset Selection ID'] = None
        else:
            self.definition['Asset Selection ID'] = value['ID']

    @property
    def html(self):
        return f'<a href="/api/content/{self.id}/sourceUrl" rel="nofollow noopener noreferrer"> ' \
               f'<img data-seeq-content="{self.id}" class="report-image-border fr-fic fr-dii contentLoaded ' \
               f'fr-draggable" src="/api/content/{self.id}/image"></a>'


class DateRange(Item):
    """
    The SPy representation of a Seeq date range. Date ranges can be created with a number of different input
    properties that a user can define when creating a Date Range asset. Date ranges can either be static,
    specified with just a Name, Start, and End, or they can be live, denoted by including Auto Enabled,
    Auto Duration, Auto offset, and Auto Offset Direction values.

    ===================== =============================================
    Input Column          Date Range Attribute
    ===================== =============================================
    ID                    The id of the date range. If not provided one
                          will be generated
    Name                  The name of the date range. Eg "Date Range 1"
    Start                 The ISO 8601 string or datetime object start
                          of the date range
    End                   The ISO 8601 string or datetime object end of
                          the date range
    Auto Enabled          Boolean if automatic update is enabled
    Auto Duration         The duration of the automatic update sliding
                          window. Eg, 10min, 1hr, 1d, etc
    Auto Offset           The offset of the automatic update sliding
                          window. Eg, 10min, 1day, etc
    Auto Offset Direction The direction of the offset. Either 'Past' or
                          'Future'. Default 'Past'
    Condition ID          The id of the condition defining the date
                          range, if applicable
    """

    DATE_RANGE_COLUMN_NAMES = ['ID', 'Name', 'Start', 'End', 'Auto Enabled', 'Auto Duration', 'Auto Offset',
                               'Auto Offset Direction', 'Condition ID', 'Type']

    def __init__(self, definition, report):
        super().__init__(definition)
        self.report = report

    def push(self, session: Session, item_map: dict, existing_date_ranges: dict):
        date_range_input = self._create_date_range_input(session, item_map)

        existing_date_range = existing_date_ranges.get(self.id)
        if not existing_date_range:
            existing_date_range = existing_date_ranges.get(self.name)

        content_api = ContentApi(session.client)
        if not existing_date_range:
            date_range_output = content_api.create_date_range(body=date_range_input)  # type: DateRangeOutputV1
        else:
            date_range_output = content_api.update_date_range(
                id=existing_date_range.id, body=date_range_input)  # type: DateRangeOutputV1

        item_map[self.id.upper()] = date_range_output.id.upper()

        return date_range_output

    STATIC_DATE_RANGE_REGEX = r'capsule\(\d+.*\d+.*\)'
    AUTO_DATE_RANGE_REGEX = r'capsule\(\$now\s*(?P<dir>[\-+])\s*(?P<offset>[\d\w\.]+)\s*-\s*(?P<dur>[\d\.]+)ms,.*\)'

    @staticmethod
    def pull(item_id, *, allowed_types=None, report=None, session: Session = None):
        session = Session.validate(session)
        if isinstance(item_id, DateRangeOutputV1):
            # This is an optimization because get_contents_with_all_metadata returns both the content and date range
            date_range_output = item_id
        else:
            content_api = ContentApi(session.client)
            date_range_output = content_api.get_date_range(id=item_id)

        date_range_dict = dict()
        date_range_dict['Name'] = date_range_output.name
        date_range_dict['ID'] = date_range_output.id
        date_range_dict['Enabled'] = date_range_output.enabled
        date_range_dict['Archived'] = date_range_output.archived

        if date_range_output.condition:
            date_range_dict['Condition ID'] = date_range_output.condition.id

        is_static_date_range = False

        if re.search(DateRange.STATIC_DATE_RANGE_REGEX, date_range_output.formula):
            is_static_date_range = True

        if is_static_date_range:
            date_range_dict['Start'] = \
                _login.parse_datetime_with_timezone(session, date_range_output.date_range.start).isoformat()
            date_range_dict['End'] = \
                _login.parse_datetime_with_timezone(session, date_range_output.date_range.end).isoformat()
        else:
            match = re.search(DateRange.AUTO_DATE_RANGE_REGEX, date_range_output.formula)
            if not match:
                raise SPyRuntimeError(f'DateRange formula not recognized: {date_range_output.formula}')
            date_range_dict['Auto Enabled'] = True
            date_range_dict['Auto Offset Direction'] = 'Past' if match.group('dir') else 'Future'
            date_range_dict['Auto Offset'] = match.group('offset')
            date_range_dict['Auto Duration'] = str(int(match.group('dur')) / 1000) + 's'

        return DateRange(date_range_dict, report)

    def _create_date_range_input(self, session: Session, item_map):
        self._validate_fields_before_push()

        # We put a dummy cron schedule here so that we can create a live date range prior to creating a report with a
        # schedule
        cron_schedule = ['59 59 23 31 12 ? 2099'] if self._is_date_range_live(session) else None

        return DateRangeInputV1(name=self.definition['Name'],
                                formula=self.get_formula(session),
                                cron_schedule=cron_schedule,
                                report_id=item_map[self.report.id],
                                enabled=_common.get(self.definition, 'Enabled', True),
                                archived=_common.get(self.definition, 'Archived', False),
                                condition_id=_common.get(self.definition, 'Condition ID'))

    def _validate_fields_before_push(self):
        if 'Name' not in self.definition:
            raise SPyValueError('Unable to push Date Range: missing "name"')

    def _is_date_range_live(self, session: Session):
        return '$now' in self.get_formula(session)

    @staticmethod
    def _validate_user_date_range(date_range):
        errors = list()

        for k in date_range.keys():
            if k not in DateRange.DATE_RANGE_COLUMN_NAMES:
                errors.append(
                    f'Unrecognized Date Range property "{k}". Valid properties:\n'
                    f'{DateRange.DATE_RANGE_COLUMN_NAMES}')

        if 'Name' not in date_range:
            errors.append('All date ranges require a "Name"')

        if 'Start' in date_range and 'End' not in date_range:
            errors.append('Date Range "End" must be supplied with Date Range "Start"')
        elif 'End' in date_range and 'Start' not in date_range:
            errors.append('Date Range "Start" must be supplied with Date Range "End"')

        # if auto enabled is defined, all the other auto fields must be defined.
        if 'Auto Enabled' in date_range:
            if 'Auto Offset' not in date_range:
                errors.append('"Auto Offset" is required if "Auto Enabled" is True')
            if 'Auto Duration' not in date_range:
                errors.append('"Auto Duration" is required if "Auto Enabled" is True')
            if 'Auto Offset Direction' not in date_range:
                errors.append('"Auto Offset Direction" is required if "Auto Enabled" is True')

        if errors:
            msg = f'There was 1 error ' if len(errors) == 1 else f'There were {len(errors)} errors '
            msg += f'detected in date range "{date_range["Name"]}": {errors}'
            raise SPyRuntimeError(msg)

    def construct_data_id(self, label):
        return self._construct_data_id(label)

    @property
    def start(self):
        return self.get_start(spy.session)

    def get_start(self, session: Session):
        return _login.parse_datetime_with_timezone(session, _common.get(self.definition, 'Start'))

    @property
    def end(self):
        return self.get_end(spy.session)

    def get_end(self, session: Session):
        return _login.parse_datetime_with_timezone(session, _common.get(self.definition, 'End'))

    def get_formula(self, session: Session):
        if 'Start' in self and 'End' in self:
            start = self.get_start(session)
            end = self.get_end(session)

            return f'capsule({int(start.value / 1_000_000)}ms,{int(end.value / 1_000_000)}ms)'
        else:
            auto_offset_direction = _common.get(self, 'Auto Offset Direction', '').lower()
            offset_direction = '-' if auto_offset_direction == 'past' else '+'
            offset_value = int(_common.parse_str_time_to_ms(self['Auto Offset'])[0])
            offset_units = _common.parse_str_time_to_ms(self['Auto Offset'])[1]
            offset_duration = int(_common.parse_str_time_to_ms(self['Auto Duration'])[2])

            return f'capsule($now {offset_direction} {offset_value}{offset_units} - {offset_duration}ms, ' \
                   f'$now {offset_direction} {offset_value}{offset_units})'


class AssetSelection(Item):
    """
    The SPy representation of a Seeq asset selection.

    ===================== =============================================
    Input Column          Asset Selection Attribute
    ===================== =============================================
    ID                    The id of the asset selection. If not
                          provided one will be generated
    Name                  The name of the asset selection. Eg "Asset
                          Selection 1"
    Asset ID              The id of the asset selected
    Path Levels           Number of levels shown in the asset path for
                          the asset selection
    """

    ASSET_SELECTION_COLUMN_NAMES = ['ID', 'Name', 'Asset ID', 'Path Levels', 'Type']

    def __init__(self, definition, report):
        super().__init__(definition)
        self.report = report

    def push(self, session: Session, item_map: dict, existing_asset_selections: dict):
        asset_selection_input = self._create_asset_selection_input(item_map)

        existing_asset_selection = existing_asset_selections.get(self.id)
        if not existing_asset_selection:
            existing_asset_selection = existing_asset_selections.get(self.name)

        content_api = ContentApi(session.client)
        if not existing_asset_selection:
            asset_selection_output = content_api.create_asset_selection(
                body=asset_selection_input)  # type: AssetSelectionOutputV1
        else:
            asset_selection_output = content_api.update_asset_selection(
                id=existing_asset_selection.id, body=asset_selection_input)  # type: AssetSelectionOutputV1

        item_map[self.id.upper()] = asset_selection_output.id.upper()

        return asset_selection_output

    @staticmethod
    def pull(item_id, *, allowed_types=None, report=None, session: Session = None):
        session = Session.validate(session)
        if isinstance(item_id, AssetSelectionOutputV1):
            asset_selection_output = item_id
        else:
            content_api = ContentApi(session.client)
            asset_selection_output = content_api.get_asset_selection(id=item_id)

        asset_selection_dict = dict()
        asset_selection_dict['Name'] = asset_selection_output.name
        asset_selection_dict['ID'] = asset_selection_output.id
        asset_selection_dict['Asset ID'] = asset_selection_output.asset.id
        asset_selection_dict['Path Levels'] = asset_selection_output.asset_path_depth
        asset_selection_dict['Archived'] = asset_selection_output.archived

        return AssetSelection(asset_selection_dict, report)

    def _create_asset_selection_input(self, item_map):
        self._validate_fields_before_push()

        return AssetSelectionInputV1(name=self.definition['Name'],
                                     selection_id=self.definition['ID'],
                                     asset_id=self.definition['Asset ID'],
                                     asset_path_depth=self.definition['Path Levels'],
                                     report_id=item_map[self.report.id],
                                     archived=_common.get(self.definition, 'Archived', False))

    def _validate_fields_before_push(self):
        if 'Name' not in self.definition:
            raise SPyValueError('Unable to push Asset Selection: missing "Name"')
        if 'Asset ID' not in self.definition:
            raise SPyValueError('Unable to push Asset Selection: missing "Asset ID"')
        if 'Path Levels' not in self.definition:
            raise SPyValueError('Unable to push Asset Selection: missing "Path Levels"')

    def construct_data_id(self, label):
        return self._construct_data_id(label)



"""
Short for Seeq PYthon, the SPy library provides methods to interact with data that is exposed to the Seeq Server.
"""
from typing import Optional

from seeq.sdk import ApiClient
from seeq.sdk.configuration import Configuration
from seeq.sdk.models import UserOutputV1
from seeq.spy import addons
from seeq.spy import assets
from seeq.spy import docs
from seeq.spy import jobs
from seeq.spy import utils
from seeq.spy import widgets
from seeq.spy import workbooks
from seeq.spy._common import PATH_ROOT, DEFAULT_WORKBOOK_PATH, GLOBALS_ONLY, GLOBALS_AND_ALL_WORKBOOKS
from seeq.spy._config import Options
from seeq.spy._login import login, logout
from seeq.spy._plot import plot
from seeq.spy._pull import pull
from seeq.spy._push import push
from seeq.spy._search import search
from seeq.spy._session import Session
from seeq.spy._status import Status

# noinspection DuplicatedCode
session: Session = Session(client_configuration=Configuration())
options: Options = session.options
client: Optional[ApiClient] = None
user: Optional[UserOutputV1] = None
server_version: Optional[str] = None

__all__ = ['addons', 'assets', 'docs', 'workbooks', 'widgets', 'login', 'logout', 'plot', 'pull', 'push', 'search',
           'PATH_ROOT', 'DEFAULT_WORKBOOK_PATH', 'GLOBALS_ONLY', 'GLOBALS_AND_ALL_WORKBOOKS', 'Session', 'Status',
           'options', 'session', 'client', 'user', 'server_version', 'jobs', 'utils']

__version__ = '%d.%d.%d.%d.%d' % (int('$SQ(VERSION_MAJOR)'), int('$SQ(VERSION_MINOR)'), int('$SQ(VERSION_PATCH)'),
                                  int('$SQ(VERSION_MAJOR_PYPI)'), int('$SQ(VERSION_MINOR_PYPI)'))
